perl -I../lib add.t 1
perl -I../lib cache.t 1
perl -I../lib omit.t 1
perl -I../lib read_file.t 1
perl -I../lib std.t 1
perl -I../lib tagchar.t 1
perl -I../lib valid_ele.t 1
perl -I../lib valid_file.t 1
